# Autogenerated by nbdev

d = { 'settings': { 'branch': 'main',
                'doc_baseurl': '/ptyp',
                'doc_host': 'https://dsm-72.github.io',
                'git_url': 'https://github.com/dsm-72/ptyp',
                'lib_path': 'ptyp'},
  'syms': { 'ptyp.cons': {},
            'ptyp.grds': { 'ptyp.grds.aliascheck': ('grds.html#aliascheck', 'ptyp/grds.py'),
                           'ptyp.grds.checknan': ('grds.html#checknan', 'ptyp/grds.py'),
                           'ptyp.grds.isallsame': ('grds.html#isallsame', 'ptyp/grds.py'),
                           'ptyp.grds.isiter': ('grds.html#isiter', 'ptyp/grds.py'),
                           'ptyp.grds.isnan': ('grds.html#isnan', 'ptyp/grds.py'),
                           'ptyp.grds.isnanlike': ('grds.html#isnanlike', 'ptyp/grds.py'),
                           'ptyp.grds.isnotiterstr': ('grds.html#isnotiterstr', 'ptyp/grds.py'),
                           'ptyp.grds.isorigin': ('grds.html#isorigin', 'ptyp/grds.py'),
                           'ptyp.grds.isstrbytes': ('grds.html#isstrbytes', 'ptyp/grds.py'),
                           'ptyp.grds.istype': ('grds.html#istype', 'ptyp/grds.py'),
                           'ptyp.grds.istyping': ('grds.html#istyping', 'ptyp/grds.py'),
                           'ptyp.grds.trychecks': ('grds.html#trychecks', 'ptyp/grds.py')},
            'ptyp.meta': { 'ptyp.meta.Alias': ('meta.html#alias', 'ptyp/meta.py'),
                           'ptyp.meta.AliasMeta': ('meta.html#aliasmeta', 'ptyp/meta.py'),
                           'ptyp.meta.AliasMeta.getbtype': ('meta.html#aliasmeta.getbtype', 'ptyp/meta.py'),
                           'ptyp.meta.Not': ('meta.html#not', 'ptyp/meta.py'),
                           'ptyp.meta.NotMeta': ('meta.html#notmeta', 'ptyp/meta.py'),
                           'ptyp.meta.NotMeta.check': ('meta.html#notmeta.check', 'ptyp/meta.py'),
                           'ptyp.meta.NotMeta.default_kwds': ('meta.html#notmeta.default_kwds', 'ptyp/meta.py'),
                           'ptyp.meta.NotMeta.getbtype': ('meta.html#notmeta.getbtype', 'ptyp/meta.py'),
                           'ptyp.meta.Opt': ('meta.html#opt', 'ptyp/meta.py'),
                           'ptyp.meta.OptionalMeta': ('meta.html#optionalmeta', 'ptyp/meta.py'),
                           'ptyp.meta.OptionalMeta.__new__': ('meta.html#optionalmeta.__new__', 'ptyp/meta.py'),
                           'ptyp.meta.OptionalMeta.default_kwds': ('meta.html#optionalmeta.default_kwds', 'ptyp/meta.py'),
                           'ptyp.meta.OptionalMeta.getbtype': ('meta.html#optionalmeta.getbtype', 'ptyp/meta.py'),
                           'ptyp.meta.ProtoType': ('meta.html#prototype', 'ptyp/meta.py'),
                           'ptyp.meta.ProtocolBaseMeta': ('meta.html#protocolbasemeta', 'ptyp/meta.py'),
                           'ptyp.meta.ProtocolBaseMeta.__call__': ('meta.html#protocolbasemeta.__call__', 'ptyp/meta.py'),
                           'ptyp.meta.ProtocolBaseMeta.__init_subclass__': ('meta.html#protocolbasemeta.__init_subclass__', 'ptyp/meta.py'),
                           'ptyp.meta.ProtocolBaseMeta.__instancecheck__': ('meta.html#protocolbasemeta.__instancecheck__', 'ptyp/meta.py'),
                           'ptyp.meta.ProtocolBaseMeta.__new__': ('meta.html#protocolbasemeta.__new__', 'ptyp/meta.py'),
                           'ptyp.meta.ProtocolBaseMeta.check': ('meta.html#protocolbasemeta.check', 'ptyp/meta.py'),
                           'ptyp.meta.ProtocolBaseMeta.classname': ('meta.html#protocolbasemeta.classname', 'ptyp/meta.py'),
                           'ptyp.meta.ProtocolBaseMeta.default_kwds': ('meta.html#protocolbasemeta.default_kwds', 'ptyp/meta.py'),
                           'ptyp.meta.ProtocolBaseMeta.getallattrs': ('meta.html#protocolbasemeta.getallattrs', 'ptyp/meta.py'),
                           'ptyp.meta.ProtocolBaseMeta.getattrs': ('meta.html#protocolbasemeta.getattrs', 'ptyp/meta.py'),
                           'ptyp.meta.ProtocolBaseMeta.getbtype': ('meta.html#protocolbasemeta.getbtype', 'ptyp/meta.py'),
                           'ptyp.meta.ProtocolBaseMeta.getguards': ('meta.html#protocolbasemeta.getguards', 'ptyp/meta.py'),
                           'ptyp.meta.ProtocolBaseMeta.gethasattrs': ('meta.html#protocolbasemeta.gethasattrs', 'ptyp/meta.py'),
                           'ptyp.meta.ProtocolBaseMeta.getmapattrs': ('meta.html#protocolbasemeta.getmapattrs', 'ptyp/meta.py'),
                           'ptyp.meta.ProtocolBaseMeta.getqualname': ('meta.html#protocolbasemeta.getqualname', 'ptyp/meta.py'),
                           'ptyp.meta.ProtocolBaseMeta.gettypes': ('meta.html#protocolbasemeta.gettypes', 'ptyp/meta.py'),
                           'ptyp.meta.ProtocolBaseMeta.makebases': ('meta.html#protocolbasemeta.makebases', 'ptyp/meta.py'),
                           'ptyp.meta.ProtocolBaseMeta.makeguard': ('meta.html#protocolbasemeta.makeguard', 'ptyp/meta.py'),
                           'ptyp.meta.ProtocolBaseMeta.metadict': ('meta.html#protocolbasemeta.metadict', 'ptyp/meta.py'),
                           'ptyp.meta.ProtocolBaseMeta.qualname': ('meta.html#protocolbasemeta.qualname', 'ptyp/meta.py'),
                           'ptyp.meta.ProtocolBaseMeta.setallattrs': ('meta.html#protocolbasemeta.setallattrs', 'ptyp/meta.py'),
                           'ptyp.meta.ProtocolBaseMeta.setattrs': ('meta.html#protocolbasemeta.setattrs', 'ptyp/meta.py'),
                           'ptyp.meta.ProtocolBaseMeta.setbtype': ('meta.html#protocolbasemeta.setbtype', 'ptyp/meta.py'),
                           'ptyp.meta.ProtocolBaseMeta.setguards': ('meta.html#protocolbasemeta.setguards', 'ptyp/meta.py'),
                           'ptyp.meta.ProtocolBaseMeta.sethasattrs': ('meta.html#protocolbasemeta.sethasattrs', 'ptyp/meta.py'),
                           'ptyp.meta.ProtocolBaseMeta.setmapattrs': ('meta.html#protocolbasemeta.setmapattrs', 'ptyp/meta.py'),
                           'ptyp.meta.ProtocolBaseMeta.setqualname': ('meta.html#protocolbasemeta.setqualname', 'ptyp/meta.py'),
                           'ptyp.meta.ProtocolBaseMeta.settypes': ('meta.html#protocolbasemeta.settypes', 'ptyp/meta.py'),
                           'ptyp.meta.ProtocolMeta': ('meta.html#protocolmeta', 'ptyp/meta.py'),
                           'ptyp.meta.ProtocolMeta.getbtype': ('meta.html#protocolmeta.getbtype', 'ptyp/meta.py')},
            'ptyp.ptyp': { 'ptyp.ptyp.IntQ': ('ptyp.html#intq', 'ptyp/ptyp.py'),
                           'ptyp.ptyp.NotStrBytes': ('ptyp.html#notstrbytes', 'ptyp/ptyp.py'),
                           'ptyp.ptyp.StrBytes': ('ptyp.html#strbytes', 'ptyp/ptyp.py'),
                           'ptyp.ptyp.StrQ': ('ptyp.html#strq', 'ptyp/ptyp.py')},
            'ptyp.type': {},
            'ptyp.util': { 'ptyp.util.argname': ('util.html#argname', 'ptyp/util.py'),
                           'ptyp.util.argsname': ('util.html#argsname', 'ptyp/util.py'),
                           'ptyp.util.classname': ('util.html#classname', 'ptyp/util.py'),
                           'ptyp.util.classvars': ('util.html#classvars', 'ptyp/util.py'),
                           'ptyp.util.filtkwds': ('util.html#filtkwds', 'ptyp/util.py'),
                           'ptyp.util.getname': ('util.html#getname', 'ptyp/util.py'),
                           'ptyp.util.guardtype': ('util.html#guardtype', 'ptyp/util.py'),
                           'ptyp.util.opttypes': ('util.html#opttypes', 'ptyp/util.py'),
                           'ptyp.util.popget': ('util.html#popget', 'ptyp/util.py'),
                           'ptyp.util.qualname': ('util.html#qualname', 'ptyp/util.py'),
                           'ptyp.util.unionify': ('util.html#unionify', 'ptyp/util.py')}}}



# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/03_util.ipynb.

# %% auto 0
__all__ = ['filtkwds', 'popget', 'classvars', 'getname', 'argname', 'argsname', 'classname', 'unionify', 'opttypes', 'guardtype',
           'qualname']

# %% ../nbs/03_util.ipynb 6
import inspect

# %% ../nbs/03_util.ipynb 8
from types import NoneType, FunctionType
from typing import (
    Self, Type, TypeVar, TypeGuard, TypeAlias, ParamSpec,
    Callable, Iterable, Optional, Protocol, _ProtocolMeta, 
    Any, Set, Dict, List, Tuple, Union, ClassVar, Generic, Literal,
    get_args, get_origin, runtime_checkable, overload, 
)

# %% ../nbs/03_util.ipynb 10
#| export


# %% ../nbs/03_util.ipynb 12
#| export


# %% ../nbs/03_util.ipynb 14
from .cons import ALLATTRS
from ptyp.type import (
    T, P, Types, TypeLike, Guard,
    TGuard, IterT, IterFunc, NanGuard, IterGuard, IterTGuard,
    IterStr, AttrGuards
)

# %% ../nbs/03_util.ipynb 16
def filtkwds(fn: Callable, **kwargs):
    sig = inspect.signature(fn).parameters
    return {k: v for k, v in kwargs.items() if k in sig}

def popget(o, attr: str, default = None, **kwargs):
    return kwargs.pop(attr, getattr(o, attr, default))

def classvars(cls: type, *attrs: str, **kwargs) -> dict:
    return dict((attr, popget(cls, attr, None, **kwargs)) for attr in attrs)

def getname(o: object) -> str: 
    return str(getattr(o, '__name__', o))

def argname(a: T) -> str: 
    return getname((get_args(a) or (a, ))[0])

def argsname(*args) -> str: 
    return ', '.join(list(map(argname, args)))

def classname(cls: type, alt: str = None) -> str:
    if inspect.isclass(alt): return getname(alt)
    if isinstance(alt, str): return alt
    return getname(cls)

def unionify(*t: Types) -> type:
    return None if (l := len(t)) == 0 else t[0] if l == 1 else Union[t[0], *t[1:]]

def opttypes(types: tuple[type], dropnone: bool = True) -> type:
    return types if not dropnone else tuple(Optional[t] for t in types if t is not None)

def guardtype(*types: Types) -> Guard:
    return TypeGuard[unionify(*opttypes(types))]

def qualname(
    cls, *types: Types, prefix: str = '', suffix: str = '', 
    dropnone: bool = True, alt: str = None,
) -> str:
    '''Generates a qualified name for a type or a set of types.'''
    types = opttypes(types, dropnone)
    cname = classname(cls, alt)
    qname = f'{prefix}{{{argsname(*types)}}}{suffix}'    
    qname = f'{cname}({qname})'
    return qname



# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/02_grds.ipynb.

# %% auto 0
__all__ = ['isnan', 'isnanlike', 'checknan', 'isiter', 'isstrbytes', 'isnotiterstr', 'isorigin', 'istyping', 'istype',
           'trychecks', 'isallsame', 'aliascheck']

# %% ../nbs/02_grds.ipynb 6
import os, io, inspect
from math import isnan as pyisnan, nan as pynan
from operator import eq, is_
from contextlib import redirect_stderr

# %% ../nbs/02_grds.ipynb 8
from types import NoneType, FunctionType
from typing import (
    Self, Type, TypeVar, TypeGuard, TypeAlias, ParamSpec,
    Callable, Iterable, Optional, Protocol, _ProtocolMeta, 
    Any, Set, Dict, List, Tuple, Union, ClassVar, Generic, Literal,
    get_args, get_origin, runtime_checkable, overload, 
)

# %% ../nbs/02_grds.ipynb 10
#| export


# %% ../nbs/02_grds.ipynb 12
#| export


# %% ../nbs/02_grds.ipynb 14
from .cons import ALLATTRS
from ptyp.type import (
    T, P, Types, TypeLike, Guard,
    TGuard, IterT, IterFunc, NanGuard, IterGuard, IterTGuard,
    IterStr, AttrGuards
)

# %% ../nbs/02_grds.ipynb 17
def isnan(o: object, t: type | None = None) -> NanGuard:
    # if t is not None and not isnan(t, None): return False # NOTE: only apply isnan check if t is None
    try: return pyisnan(o)
    except: return False
    
def isnanlike(o: object, t: type) -> NanGuard:
    return all((isnan(t), isnan(o)))

def checknan(o: object, t: type) -> NanGuard:
    if isnan(t): return isnan(o)
    return False

# %% ../nbs/02_grds.ipynb 19
def isiter(o: object) -> IterGuard: 
    return isinstance(o, Iterable)

def isstrbytes(o: object) -> TypeGuard[Union[str, bytes]]: 
    return isinstance(o, (str, bytes))

def isnotiterstr(x) -> IterGuard: 
    return isiter(x) and not isstrbytes(x)

# %% ../nbs/02_grds.ipynb 21
def isorigin(t: T) -> TypeGuard[Union[Set, List, type, Type, Tuple, Iterable]]:
    types = (Set, List, type, Type, Tuple, Union, Literal, Iterable)
    return any((get_origin(t) is _ for _ in types))

def istyping(o: object, t: T) -> TypeGuard[T]:
    if not isorigin(t): return False
    try: return any(((o == (args := get_args(t))), o in args, o == t))
    except: return False
    
def istype(o: object, t: T, guards: IterFunc = tuple()) -> TGuard:
    funcs = (isinstance, istyping, eq, is_) + tuple(guards)
    with redirect_stderr(io.StringIO()):
        for comp in funcs:
            try:
                if comp(o, t): return True
            except: continue
    return False


# %% ../nbs/02_grds.ipynb 23
def trychecks(o: object, t: type) -> Guard:
    fns = (isinstance, issubclass, eq)
    if isnan(t): fns += (checknan, )
    for f in fns:
        try:
            if f(o, t): return True
        except: continue
    return False

# %% ../nbs/02_grds.ipynb 25
def isallsame(it: Iterable, dtype: T) -> IterTGuard:
    '''Checks if all elements in an iterable are of the same specified type.'''
    # not iterable, confirm type
    if not isiter(it): return trychecks(it, dtype)
    for el in it:
        # check if element is iterable and not a string (e.g. list, array, tuple, etc)
        if isnotiterstr(el): # if iterable, then recurse each element in the iterable
            try:                
                if not isallsame(el, dtype): return False 
            except TypeError: return False
            
        elif not trychecks(el, dtype): return False # otherwise check just this element
    return True

# %% ../nbs/02_grds.ipynb 27
def aliascheck(
    obj: T,           # object to check
    types: IterT, # types to check
    guards: IterFunc = tuple(),
    attrs: dict = dict(),
    hasattrs: IterStr = tuple(),  # attributes to check
    allattrs: IterStr = ALLATTRS, # attributes to check that values are all the same if object is iterable
    mapattrs: AttrGuards = dict(),
):
    '''Checks if an object matches a set of types and specific attributes.'''
    # print('aliascheck', obj, types, guards, attrs, hasattrs, allattrs, mapattrs)
    hasattrs = tuple(hasattrs) + tuple(k for k, v in attrs.items() if v not in allattrs)
    if not all(hasattr(obj, attr) for attr in hasattrs): return False
    
    
    found = any(istype(obj, t, guards) for t in types)
    # print(obj, found, list(istype(obj, t, guards) for t in types))
    if not found: return False
    
    for attr, val in attrs.items():        
        if getattr(obj, attr, None) == val:
            continue
        if attr in allattrs and isnotiterstr(obj) and isallsame(obj, val): 
            continue
        if attr in mapattrs and mapattrs[attr](obj, val):
            continue
    return True



# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/00_cons.ipynb.

# %% auto 0
__all__ = ['ALLATTRS']

# %% ../nbs/00_cons.ipynb 6
from math import nan as pynan

# %% ../nbs/00_cons.ipynb 8
from types import NoneType, FunctionType
from typing import (
    Self, Type, TypeVar, TypeGuard, TypeAlias, ParamSpec,
    Callable, Iterable, Optional, Protocol, _ProtocolMeta, 
    Any, Set, Dict, List, Tuple, Union, ClassVar, Generic, Literal,
    get_args, get_origin, runtime_checkable, overload, 
)

# %% ../nbs/00_cons.ipynb 10
#| export


# %% ../nbs/00_cons.ipynb 12
#| export


# %% ../nbs/00_cons.ipynb 14
#| export


# %% ../nbs/00_cons.ipynb 16
ALLATTRS = {'types', 'dtype', 'type'}
